"""
_LinuxUSBDetector: This platform-specific implementation of the _USBDetectorBase class is designed for Linux systems.
It provides the necessary functionality to detect USB devices connected to a Linux system and monitor changes in their
connections. The class utilizes the pyudev library to interact with the Linux system's device management
subsystem (udev).

Author: Eric-Canas
Date: 28-03-2023
Email: eric@ericcanas.com
Github: https://github.com/Eric-Canas
"""

from __future__ import annotations

from ._constants import _SECONDS_BETWEEN_CHECKS, _LINUX_TUPLE_ATTRIBUTES_SEPARATORS
from ..attributes import ID_VENDOR_ID, DEVTYPE, DEVICE_ATTRIBUTES, DEVNAME

from ._usb_detector_base import _USBDetectorBase


class _LinuxUSBDetector(_USBDetectorBase):
    def __init__(self, filter_devices: list[dict[str, str]] | tuple[dict[str, str]] | None = None):
        import pyudev
        self.context = pyudev.Context()
        self.monitor = None
        super(_LinuxUSBDetector, self).__init__(filter_devices=filter_devices)

    def get_available_devices(self) -> dict[str, dict[str, str | tuple[str, ...]]]:
        """
        Returns a dictionary of the currently available devices, where the key is the device ID and the value is a
        dictionary of the device's information.
        :return: dict[str, dict[str, str]]. The key is the device ID, the value is a dictionary of the device's
                information.
        """
        usb_devices = [device for device in self.context.list_devices(subsystem='usb') if ID_VENDOR_ID in device]
        devices_info = {}
        for device in usb_devices:
            device_id = device[DEVNAME]
            device_info = {attr: device.get(attr, "") for attr in DEVICE_ATTRIBUTES}
            device_info = self.__generate_tuple_attributes_from_string(device_info=device_info)
            devices_info[device_id] = device_info

        if self.filter_devices is not None:
            devices_info = self._apply_devices_filter(devices=devices_info)

        return devices_info

    def __generate_tuple_attributes_from_string(self, device_info: dict[str, str]) -> dict[str, tuple[str]|str]:
        """
        Generates a tuple of attributes for those attributes that are expected to be a tuple,
        but are stored as a string.
        :param device_info: dict[str, str]. The device information.
        :return: dict[str, tuple[str]|str]. The device information with the tuple attributes.
        """
        for attribute, separator in _LINUX_TUPLE_ATTRIBUTES_SEPARATORS.items():
            if attribute in device_info:
                assert isinstance(device_info[attribute], str), f"The attribute '{attribute}' is expected to be a string"
                # noinspection PyTypeChecker
                device_info[attribute] = tuple(value for value in device_info[attribute].split(separator) if value != "")
        return device_info

    def _monitor_changes(self, on_connect: callable | None = None, on_disconnect: callable | None = None,
                         check_every_seconds: int | float = _SECONDS_BETWEEN_CHECKS) -> None:
        import pyudev

        def __handle_device_event(device):
            action = device.action
            if device.get(DEVTYPE) == 'usb_device':
                device_id = device[DEVNAME]
                if action == "add" and on_connect is not None:
                    device_info = {attr: device.get(attr, "") for attr in DEVICE_ATTRIBUTES}
                    device_info = self.__generate_tuple_attributes_from_string(device_info=device_info)
                    on_connect(device_id, device_info)
                    self.last_check_devices = self.get_available_devices()
                elif action == "remove" and on_disconnect is not None and device_id in self.last_check_devices:
                    device_info = self.last_check_devices[device_id].copy()
                    on_disconnect(device_id, device_info)
                    self.last_check_devices = self.get_available_devices()

        if self.monitor is None:
            self.monitor = pyudev.Monitor.from_netlink(self.context)
            self.monitor.filter_by(subsystem='usb')

        self._thread = pyudev.MonitorObserver(self.monitor, name="USB Monitor", callback=__handle_device_event)

        # Start the observer thread
        self._thread.start()

        # Keep the main thread alive, checking for changes every specified interval
        while not self._stop_thread.is_set():
            self._stop_thread.wait(check_every_seconds)

        # Stop the observer thread
        self._thread.stop()


from setuptools import setup, find_packages

setup(
    name='usb-monitor',
    version='1.23',
    author='Eric-Canas',
    author_email='eric@ericcanas.com',
    url='https://github.com/Eric-Canas/USBMonitor',
    description='USBMonitor is an easy-to-use cross-platform library for USB device monitoring that simplifies '
                'tracking of connections, disconnections, and examination of connected device attributes on '
                'Windows, Linux and MacOs, freeing the user from platform-specific details or incompatibilities.',

    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    license='MIT',
    packages=find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'pyudev; platform_system=="Linux"',
        'pywin32; platform_system=="Windows"',
        'wmi; platform_system=="Windows"',
    ],

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Intended Audience :: Developers',
        'Topic :: System :: Hardware',
        'Topic :: System :: Hardware :: Hardware Drivers',
        'Topic :: System :: Systems Administration',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Operating System',
        'Topic :: Software Development :: Embedded Systems',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
)


from .usbmonitor import USBMonitor


# Import platform-specific detectors
from .__platform_specific_detectors._windows_usb_detector import _WindowsUSBDetector
from .__platform_specific_detectors._linux_usb_detector import _LinuxUSBDetector


"""
This file contains constant values and mappings used in the platform-specific USB device detection implementations.
It includes constants for attributes, attribute separators, and regex patterns for specific Operating Systems.
Additionally, it defines the default time interval for checking USB device changes.

Author: Eric-Canas
Date: 28-03-2023
Email: eric@ericcanas.com
Github: https://github.com/Eric-Canas
"""

from ..attributes import ID_MODEL_ID, ID_VENDOR, ID_MODEL, ID_VENDOR_FROM_DATABASE, ID_MODEL_FROM_DATABASE, \
    DEVNAME, ID_USB_CLASS_FROM_DATABASE, ID_USB_INTERFACES, DEVTYPE, ID_VENDOR_ID, ID_SERIAL

_SECONDS_BETWEEN_CHECKS = 0.5
_THREAD_JOIN_TIMEOUT_SECONDS = 5

_DEVICE_ID, _PNP_DEVICE_ID = 'DeviceID', 'PNPDeviceID'

_LINUX_TO_WINDOWS_ATTRIBUTES = {
    ID_MODEL_ID: 'HardwareID',
    ID_MODEL: 'Name',
    ID_MODEL_FROM_DATABASE: 'Caption',
    ID_VENDOR_ID: 'HardwareID',
    ID_VENDOR: 'Name',
    ID_VENDOR_FROM_DATABASE: 'Manufacturer',
    ID_USB_INTERFACES: 'CompatibleID',
    ID_USB_CLASS_FROM_DATABASE: 'PNPClass',
    DEVNAME: _DEVICE_ID,
    DEVTYPE: _PNP_DEVICE_ID,
    ID_SERIAL: _PNP_DEVICE_ID
}

_LINUX_TUPLE_ATTRIBUTES_SEPARATORS = {ID_USB_INTERFACES: ':'}

USB, USBSTOR, USB4, USBPRINT = 'USB', 'USBSTOR', 'USB4', 'USBPRINT'

_WINDOWS_USB_REGEX_ATTRIBUTES = {ID_MODEL_ID: r'PID_([0-9A-Fa-f]{4})', ID_VENDOR_ID: r'VID_([0-9A-Fa-f]{4})',
                                 DEVTYPE: r'^(.+?)\\', ID_SERIAL: r'\\([^\\]+)$'}
_WINDOWS_USB4_REGEX_ATTRIBUTES = {ID_MODEL_ID: r'PID_([0-9A-Fa-f]{4})', ID_VENDOR_ID: r'VID_([0-9A-Fa-f]{4})',
                                  DEVTYPE: r'^(.+?)\\', ID_SERIAL: r'\\([^\\]+)$'}
_WINDOWS_USBPRINT_REGEX_ATTRIBUTES = {ID_MODEL_ID: r'PID_([0-9A-Fa-f]{4})', ID_VENDOR_ID: r'VID_([0-9A-Fa-f]{4})',
                                        DEVTYPE: r'^(.+?)\\', ID_SERIAL: r'\\([^\\]+)$'}
_WINDOWS_USBSTOR_REGEX_ATTRIBUTES = {ID_MODEL_ID: r'PROD_([a-zA-Z0-9\_\/\.\-]{2,16})&', ID_VENDOR_ID: r'VEN_([a-zA-Z0-9\.\_\-\/]{2,8})&',
                                     DEVTYPE: r'^(.+?)\\', ID_SERIAL: r'\\([^\\]+)$'}

_WINDOWS_REGEX_ATTRIBUTES_BY_DRIVER = {USB: _WINDOWS_USB_REGEX_ATTRIBUTES,
                                       USBSTOR: _WINDOWS_USBSTOR_REGEX_ATTRIBUTES,
                                       USB4: _WINDOWS_USB4_REGEX_ATTRIBUTES,
                                       USBPRINT: _WINDOWS_USBPRINT_REGEX_ATTRIBUTES}

_WINDOWS_TO_LOWERCASE_ATTRIBUTES = (ID_MODEL_ID, ID_VENDOR_ID)
_WINDOWS_NON_USB_DEVICES_IDS = ("ROOT_HUB20", "ROOT_HUB30", "VIRTUAL_POWER_PDO")
_WINDOWS_USB_QUERY = f"SELECT {', '.join(set(_LINUX_TO_WINDOWS_ATTRIBUTES.values()))} FROM Win32_PnPEntity " \
                          f"WHERE {_PNP_DEVICE_ID} LIKE 'USB%'"

# Darwin-specific constants
_DARWIN_TO_LINUX_ATTRIBUTES = {
    ID_MODEL_ID: 'idProduct',
    ID_MODEL: 'kUSBProductString',
    ID_MODEL_FROM_DATABASE: 'USB Product Name',
    ID_VENDOR_ID: 'idVendor',
    ID_VENDOR: 'kUSBVendorString',
    ID_VENDOR_FROM_DATABASE: 'USB Vendor Name',
    ID_USB_INTERFACES: 'IOCFPlugInTypes',
    ID_USB_CLASS_FROM_DATABASE: 'bDeviceClass',
    DEVNAME: 'BSD Name',
    DEVTYPE: 'Device Speed',
    ID_SERIAL: 'kUSBSerialNumberString',
}

_DARWIN_REGEX_ATTRIBUTES = {
    ID_MODEL_ID: r'idProduct: ([0-9A-Fa-f]{4})',
    ID_VENDOR_ID: r'idVendor: ([0-9A-Fa-f]{4})'
}
